/*****************************************************************************
 *
 * FILE:	gpsdump.c
 * DESCRIPTION:	Print GPS (Global Positioning System) data (sample program)
 * DATE:	Wed, Jul  2 2003
 * UPDATED:	Sat, Sep 20 2003
 * AUTHOR:	Kouichi ABE (WALL)
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2003 Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (C) 2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: gpsdump.c,v 1.3 2003/09/21 11:16:50 kouichi Exp $
 *
 *****************************************************************************/

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <signal.h>
#if	HAVE_LIMITS_H
#include <limits.h>
#endif	/* HAVE_LIMITS_H */
#include <time.h>
#include <errno.h>
#include "griffin.h"


/*****************************************************************************
 *
 *	Program & Copyright
 *
 *****************************************************************************/
const char	Program[] = "gpsdump";
const char	Copyright[] =
  "GPSDUMP/1.0.2 (2003/09/20)\n\
Copyright (C) 2003 Kouichi ABE (WALL), All rights reserved.";


/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	DATE_SIZE	(32)	/* buffer size for date and time */


/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	usage(void);
static void	quit(void);


/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	dumpData
 * Description:		print GPS data (cooked)
 * Arguments:		none
 * Return Value:	none
 */
static void
dumpData(void)
{
  GPSInfo	gpsi;
  int		cnt = 0;

  /* send GPS data */
  while (gpsGetInfo(&gpsi) == 0) {
    char	date[DATE_SIZE];
    time_t	now = timegm(&gpsi.timestamp);
    struct tm *	jst = localtime(&now);

    memset(date, 0, sizeof(date));
    strftime(date, sizeof(date) - 1, "%a %b %d %H:%M:%S %Y %Z", jst);

    fprintf(stdout,
	"%06d,%lu,%s,%c,%02d.%02d.%02d.%04d,%lld,%c,%02d.%02d.%02d.%04d,%lld,%.1f,%.1f,%.1f\n",
		cnt,			/* ǡֹ	*/
		now,			/* դȻʿ͡ */
		date,			/* դȻ	*/
		gpsi.ns,		/* ̰ޡ	*/
		gpsi.latitude.degrees,	/* ()	*/
		gpsi.latitude.minutes,	/* (ʬ)	*/
		gpsi.latitude.seconds,	/* ()	*/
		gpsi.latitude.fraction,	/* (ü)	*/
		gpsi.latitude.value,	/* ()	*/
		gpsi.ew,		/* С	*/
		gpsi.longitude.degrees,	/* ()	*/
		gpsi.longitude.minutes,	/* (ʬ)	*/
		gpsi.longitude.seconds,	/* ()	*/
		gpsi.longitude.fraction,/* (ü)	*/
		gpsi.longitude.value,	/* ()	*/
		gpsi.velocity,		/* ®		*/
		gpsi.direction,		/* 		*/
		gpsi.altitude);		/* ɸ		*/
    fflush(stdout);

    /* count up */
    if (++cnt > INT_MAX) {
      cnt = 0;	/* reset counter */
    }
  }
}


/*****************************************************************************/

/*
 * Function Name:	usage
 * Description:		show the usage of this program 
 * Arguments:		none
 * Return Value:	none
 */
static void
usage(void)
{
  fprintf(stderr,
	  "%s\n"
	  "usage: %s -f <pacy|nmea> [-d device]\n",
	  Copyright, Program);
  exit(64);
}

static void
quit(void)
{
  gpsQuit();
  exit(0);
}

/*****************************************************************************/

/*
 * Main Function
 */
int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  GPSDataFormat	data_format = GDF_UNKNOWN;
  const char *	device = "/dev/cuaa0";
  register int	ch;

  /* processing arguments */
  if (argc == 1) {
    usage();
  }

  while ((ch = getopt(argc, argv, "f:d:")) != -1) {
    switch (ch) {
      case 'f':
	if (strncasecmp(optarg, "pacy", 4) == 0) {
  	  data_format = GDF_SONY_PACY;
   	}
    	else if (strncasecmp(optarg, "nmea", 4) == 0) {
  	  data_format = GDF_NMEA0183;
   	}
    	else {
  	  usage();
   	}
	break;
      case 'd':
	device = optarg;
	break;
      default:
	usage();
    }
  }
  argc -= optind;
  argv += optind;

  /* check arguments */
  if (data_format == GDF_UNKNOWN) {
    usage();
  }

  if (gpsInit(data_format, device) == 0) {
    struct sigaction	act;

    sigemptyset(&act.sa_mask);
    act.sa_flags   = 0;
    act.sa_handler = (void *)quit;
    /* quit */
    sigaction(SIGINT,  &act, NULL);
    sigaction(SIGQUIT, &act, NULL);
    sigaction(SIGTERM, &act, NULL);
    /* bug */
    sigaction(SIGBUS,  &act, NULL);
    sigaction(SIGSEGV, &act, NULL);

    dumpData();
  }

  return 0;
}
