/*****************************************************************************
 *
 * FILE:	pacy.c
 * DESCRIPTION:	GPS functions for SONY PACY-CNV10 (PC-Card)
 * DATE:	Thu, Dec 20 2001
 * UPDATED:	Sat, Sep 20 2003
 * AUTHOR:	Kouichi ABE (WALL)
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2001-2003 Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (C) 2001-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pacy.c,v 1.4 2003/09/21 11:16:50 kouichi Exp $
 *
 *****************************************************************************/

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include "griffin.h"
#include "gps-sub.h"


/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	INVALID_VALUE		(-999)


/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static struct tm *	str2localtm(char *);
static void		get_ch_info(struct _channel *, int);


/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	pacy_getinfo
 * Description:		get information from SONY PACY-PNV10 GPS Receiver
 * Arguments:		struct _pacy *, FILE *
 * Return Value:	Successful: 0, otherwise: -1
 */
int
pacy_getinfo(pacy, fin)
	register struct _pacy *	pacy;
	register FILE *		fin;
{
  static char	mapdatum_str[][94] = {	/* ǡ	*/
    "WGS-84",	/* ɬ̵ */
    "Tokyo",
    "ADINDAN",
    "ARC 1950",
    "MERCHICH",
    "HONG KONG 1963",
    "SOUTH ASIA",
    "LUZON",
    "INDIAN",
    "INDIAN",
    "FERTAU 1948",
    "NORTH AMERICAN 1927 (use for Belize, Costa Rica, El Salvador, "
    "Guatemala, Honduras, Nicaragua)",
    "EUROPEAN 1950, EUROPEAN 1979",
    "IRELAND 1965     ",
    "ORDNANCE SURVEY OF GREAT BRITAN 1936",
    "NAHRWAN",
    "NAHRWAN",
    "OLD EGYPTIAN",
    "NORTH AMERICAN 1927  (use for Canada, Newfoundland Island)",
    "NORTH AMERICAN 1983  (Alaska, Canada, Mexico, Central America, USA)",
    "AUSTRALIAN GEODETIC 1984",
    "GEODETIC DATUM 1949",
    "PROVISIONAL SOUTH AMERICAN 1956",
    "SOUTH AMERICAN 1969",
    "CAMPO INCHAUSPE",
    "CORREGO ALEGRE"
  };
  static char	dop_str[][8] = {
    "1", "2", "3", "4", "5", "6", "7", "8", "9", "10",
    "11 - 12", "13 - 15", "16 - 20", "21 - 30", "31 - 50",
    "51 - 99", "100+"
  };
  static int	dop_val[] = {
    1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 14, 18, 25, 40, 75, 100
  };
  char		buf[GBUFSIZE];

  if (fgets(buf, sizeof(buf), fin) != NULL) {
    char	s[GBUFSIZE];
    struct tm *	tm;
    int		i, j;

    /* vendor version */
    sprintf(pacy->version, "%6.6s", buf);

    /* current date and time */
    sprintf(s, "%13.13s", buf + 6);
    tm = str2localtm(s);
    memcpy(&(pacy->nowtm), tm, sizeof(struct tm));

    /* latitude */
    pacy->ns = buf[19];
    if (pacy->ns == 'n') {
      pacy->ns = 'N';
    }
    else if (pacy->ns == 's') {
      pacy->ns = 'S';
    }

    sprintf(s, "%2.2s", buf + 20);
    pacy->latitude_deg = atoi(s);

    sprintf(s, "%2.2s", buf + 22);
    pacy->latitude_min = atoi(s);

    sprintf(s, "%2.2s", buf + 24);
    pacy->latitude_sec = atoi(s);

    sprintf(s, "%1.1s", buf + 26);
    pacy->latitude_frac = atoi(s);

    pacy->n_latitude = pacy->latitude_deg +
		       pacy->latitude_min / 60. +
		       pacy->latitude_sec / 3600. +
		       pacy->latitude_frac / 36000.;

    if (pacy->ns == 'S') {
      pacy->n_latitude = -(pacy->n_latitude);
    }

    /* longitude */
    pacy->ew = buf[27];
    sprintf(s, "%3.3s", buf + 28);
    pacy->longitude_deg = atoi(s);

    sprintf(s, "%2.2s", buf + 31);
    pacy->longitude_min = atoi(s);

    sprintf(s, "%2.2s", buf + 33);
    pacy->longitude_sec = atoi(s);

    sprintf(s, "%1.1s", buf + 35);
    pacy->longitude_frac = atoi(s);

    sprintf(s, "%5.5s", buf + 36);
    pacy->n_longitude = pacy->longitude_deg +
			pacy->longitude_min / 60. +
			pacy->longitude_sec / 3600. +
			pacy->longitude_frac / 36000.;
    if (pacy->ew == 'W') {
      pacy->n_longitude = -(pacy->n_longitude);
    }

    /* altitude */
    pacy->altitude = atoi(s);

    /* velocity */
    sprintf(s, "%3.3s", buf + 41);
    pacy->velocity = atoi(s);

    /* direction */
    sprintf(s, "%3.3s", buf + 44);
    pacy->direction = atoi(s);

    /* record date and time ? */
    sprintf(s, "%13.13s", buf + 47);
    tm = str2localtm(s);
    memcpy(&(pacy->calctm), tm, sizeof(struct tm));

    /* dilution of precision (DOP) */
    pacy->dop	= dop_str[buf[60] - 'A'];
    pacy->n_dop = dop_val[buf[60] - 'A'];	/* non-exactly */

    /* measurement mode */
    sprintf(s, "%1.1s", buf + 61);
    pacy->measuremode = atoi(s);

    /* map datum of position data */
    pacy->mapdatum = mapdatum_str[buf[62] - 'A'];

    /* satellite information */
    for (i = 0, j = 63; i < 8; i++, j += 5) {
      sprintf(pacy->ch[i].status, "%5.5s", buf + j);
      get_ch_info(&(pacy->ch[i]), i + 1);
    }

    /* unknown data */
    sprintf(pacy->rest, "%4.4s", buf + 103);

    /* parity */
    pacy->parity = buf[107];

    return 0;
  }

  return -1;
}

/*
 * Function Name:	str2localtm
 * Description:
 * Arguments:
 * Return Value:
 */
static struct tm *
str2localtm(buf)
	register char *	buf;
{
  static int		yday[] = {0, 31, 59, 90, 120, 151, 181, 212, 242, 273,
				  303, 334};
  static struct tm	tm;
  int			year;
  int			leap;
  char			s[GBUFSIZE];

  sprintf(s, "%2.2s", buf);
  tm.tm_year	= atoi(s);
  sprintf(s, "%2.2s", buf + 2);
  tm.tm_mon	= atoi(s) - 1;
  sprintf(s, "%2.2s", buf + 4);
  tm.tm_mday	= atoi(s);
  sprintf(s, "%1.1s", buf + 6);
  tm.tm_wday	= atoi(s);
  sprintf(s, "%2.2s", buf + 7);
  tm.tm_hour	= atoi(s);
  sprintf(s, "%2.2s", buf + 9);
  tm.tm_min	= atoi(s);
  sprintf(s, "%2.2s", buf + 11);
  tm.tm_sec	= atoi(s);
  if (tm.tm_year < 70 ) {
    tm.tm_year += 100;
  }
  year	= tm.tm_year + 1900;
  leap	= ((year % 4 == 0) && (year % 100 != 0)) || (year % 400 == 0) ? 1 : 0;
  if (tm.tm_mon > 1) {
    tm.tm_yday = yday[tm.tm_mon] + tm.tm_mday + leap;
  }
  else {
    tm.tm_yday = yday[tm.tm_mon] + tm.tm_mday;
  }
  tm.tm_isdst	= 0;
  tm.tm_gmtoff	= DEF_OFFSET;
  tm.tm_zone	= DEF_TIMEZONE;

  return &tm;
}

/*
 * Function Name:	get_ch_info
 * Description:		get the information of current channel
 * Arguments:		struct _channel *, channel number
 * Return Value:	none
 */
static void
get_ch_info(ch, chno)
	register struct _channel	*	ch;
	const int			chno;
{
  static char *	prn_str = "ABCDEFGHIJKLMNOPQRSTUVWXabcdefgh";
  static char	elevation_str[][7] = {
    " 0.. 5", " 6..15", "16..25", "26..35", "36..45",
    "46..55", "56..65", "66..75", "76..85", "86..90"
  };
  static int	elevation_val[] = {
    0, 10, 20, 30, 40, 50, 60, 70, 80, 90
  };
  static char	azimuth_str[][9] = {
    "  0..  5", "  6.. 15", " 16.. 25", " 26.. 35", " 36.. 45",
    " 46.. 55", " 56.. 65", " 66.. 75", " 76.. 85", " 86.. 95",
    " 90..105", "106..115", "116..125", "126..135", "136..145",
    "146..155", "156..165", "166..175", "176..180"
  };
  static int	azimuth_val[] = {
    0, 10, 20, 30, 40, 50, 60, 70, 80, 90,
    100, 110, 120, 130, 140, 150, 160, 170, 180
  };
  static char	situation_str[][8] = {
    "SCAN", "LOCK", "READY", "HOLD", "ILL", "OK"
  };
  static char	situation_char[] = ".o+-x*";
  char *	chp;

  ch->number = chno;
  chp = strchr(prn_str, ch->status[0]);
  if (chp != NULL) {
    ch->prn = (chp - prn_str) + 1;
  }
  else {
    ch->prn = INVALID_VALUE;
  }

  if ('A' <= ch->status[1] && ch->status[1] <= 'J') {
    snprintf(ch->elevation, sizeof(ch->elevation) - 1, "  %s ",
	     elevation_str[ch->status[1] - 'A']);
    ch->n_elevation = elevation_val[ch->status[1] - 'A'];
  }
  else if ('a' <= ch->status[1] && ch->status[1] <= 'j') {
    snprintf(ch->elevation, sizeof(ch->elevation) - 1, "-(%s)",
	     elevation_str[ch->status[1] - 'a']);
    ch->n_elevation = -elevation_val[ch->status[1] - 'a'];
  }
  else {
    snprintf(ch->elevation, sizeof(ch->elevation) - 1, "???");
    ch->n_elevation = INVALID_VALUE;
  }

  if ('A' <= ch->status[2] && ch->status[2] <= 'S') {
    snprintf(ch->azimuth, sizeof(ch->azimuth) - 1, "  %s ",
	     azimuth_str[ch->status[2] - 'A']);
    ch->n_azimuth = azimuth_val[ch->status[2] - 'A'];
  }
  else if ('a' <= ch->status[2] && ch->status[2] <= 's') {
    snprintf(ch->azimuth, sizeof(ch->azimuth) - 1, "-(%s)",
	     azimuth_str[ch->status[2] - 'a']);
    ch->n_azimuth = -azimuth_val[ch->status[2] - 'a'];
  }
  else {
    sprintf(ch->azimuth, "???");
    ch->n_azimuth = INVALID_VALUE;
  }

  ch->n_situation = ch->status[3] - 'A';
  ch->situation	  = situation_str[ch->n_situation];
  ch->c_situation = situation_char[ch->n_situation];

  ch->signal = (ch->status[4] - 'A') * 4;
}

/*
 * Function Name:	pacy_raw2std
 * Description:		convert GPS raw data to standard information data
 * Arguments:		GPS {raw,standard} information structure
 * Return Value:	none
 */
void
pacy_raw2std(pacy, gpsi)
	register struct _pacy *	pacy;
	register GPSInfo *	gpsi;
{
  gpsi->timestamp		= pacy->calctm;
  gpsi->ns			= pacy->ns;		/* ̰ޡ	*/
  gpsi->latitude.value		= (pacy->latitude_deg * 3600
				+ pacy->latitude_min * 60
				+ pacy->latitude_sec) * 10
				+ pacy->latitude_frac;
  gpsi->latitude.degrees	= pacy->latitude_deg;	/* ١١	*/
  gpsi->latitude.minutes	= pacy->latitude_min;	/* ١ʬ	*/
  gpsi->latitude.seconds	= pacy->latitude_sec;	/* ١á	*/
  gpsi->latitude.fraction	= pacy->latitude_frac;	/* ١1/10á*/
  gpsi->ew			= pacy->ew;		/* С	*/
  gpsi->longitude.value		= (pacy->longitude_deg * 3600
				+ pacy->longitude_min * 60
				+ pacy->longitude_sec) * 10
				+ pacy->longitude_frac;
  gpsi->longitude.degrees	= pacy->longitude_deg;	/* ١١	*/
  gpsi->longitude.minutes	= pacy->longitude_min;	/* ١ʬ	*/
  gpsi->longitude.seconds	= pacy->longitude_sec;	/* ١á	*/
  gpsi->longitude.fraction	= pacy->longitude_frac;	/* ١1/10á*/
  gpsi->velocity		= (double)pacy->velocity;	/* ®	*/
  gpsi->direction		= (double)pacy->direction;	/* */
  gpsi->altitude		= (double)pacy->altitude;	/* ɸ	*/

  gpsi->dop			= pacy->n_dop;		/* ¬	*/
  gpsi->status			= '1';			/* ¬̾	*/
  gpsi->satellite		= pacy->measuremode;	/* ѱ	*/
}

